package dev.peerat.framework;

import java.lang.reflect.Executable;
import java.lang.reflect.Field;
import java.lang.reflect.Parameter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.BiFunction;

public class DependencyInjector{
	
	private Map<String, Object> map;
	private List<BiFunction<Injection, Class<?>, Object>> builders;
	private Object[] injections;
	private Map<Class<?>, Object> cache;
	
	public DependencyInjector(){
		this.map = new HashMap<>();
		this.builders = new ArrayList<>();
	}
	
	public DependencyInjector of(String name, Object value){
		this.map.put(name, value);
		return this;
	}
	
	public DependencyInjector of(BiFunction<Injection, Class<?>, Object> builder){
		this.builders.add(builder);
		return this;
	}
	
	public DependencyInjector of(Object... injections){
		if(this.injections == null){
			this.injections = injections;
			this.cache = new HashMap<>();
			return this;
		}
		Object[] copy = new Object[this.injections.length+injections.length];
		System.arraycopy(this.injections, 0, copy, 0, this.injections.length);
		System.arraycopy(injections, 0, copy, this.injections.length, injections.length);
		this.injections = copy;
		return this;
	}
	
	public Object[] apply(Executable executable) throws Exception{
		Parameter[] parameters = executable.getParameters();
		Object[] injections = new Object[parameters.length];
		for(int i = 0; i < parameters.length; i++) injections[i] = find(parameters[i].getAnnotation(Injection.class), parameters[i].getType());
		return injections;		
	}
	
	public Object apply(Field field) throws Exception{
		return find(field.getAnnotation(Injection.class), field.getType());
	}
	
	public Object find(Injection annotation, Class<?> type){
		if(annotation != null){
			Object result = this.map.get(annotation.value());
			if(result == null) throw new IllegalArgumentException("No depdency named "+annotation.value()+" ("+this.map.keySet()+")");
			return result;
		}
		for(BiFunction<Injection, Class<?>, Object> function : builders){
			Object result = function.apply(annotation, type);
			if(result != null) return result;
		}
		if(this.injections == null) throw new IllegalArgumentException("No dependency for type "+type);
		Object result = this.cache.get(type);
		if(result != null) return result;
		for(Object injection : injections){
			if(type.isAssignableFrom(injection.getClass())){
				if(cache.containsKey(type)) throw new IllegalArgumentException("Double dependency for type "+type);
				cache.put(type, injection);
				result = injection;
			}
		}
		if(result == null) throw new IllegalArgumentException("No dependency for type "+type);
		return result;
	}
	
	@Override
	public String toString(){
		return "([DependencyInjector] map="+map+",buildersCount="+builders.size()+",injections="+injections+",cache="+cache+")";
	}
}