package dev.peerat.framework.auth;

import java.security.Key;
import java.util.Map;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.function.Function;

import org.jose4j.jwa.AlgorithmConstraints.ConstraintType;
import org.jose4j.jwk.JsonWebKey.OutputControlLevel;
import org.jose4j.jwk.RsaJsonWebKey;
import org.jose4j.jwk.RsaJwkGenerator;
import org.jose4j.jws.AlgorithmIdentifiers;
import org.jose4j.jws.JsonWebSignature;
import org.jose4j.jwt.JwtClaims;
import org.jose4j.jwt.consumer.JwtConsumer;
import org.jose4j.jwt.consumer.JwtConsumerBuilder;

import dev.peerat.framework.HttpReader;
import dev.peerat.framework.RequestType;
import dev.peerat.framework.User;

public class JwtAuthenticator<U extends User> implements Authenticator<U>{
	
	private RsaJsonWebKey rsaJsonWebKey;
	private JwtConsumer jwtConsumer;
	private BiConsumer<JwtClaims, U> claims;
	private Function<JwtClaims, U> userCreator;
	
	public JwtAuthenticator(){}
	
	public JwtAuthenticator<U> configure(Consumer<JwtConsumerBuilder> consumer, BiConsumer<JwtClaims, U> claims, Function<JwtClaims, U> userCreator) throws Exception{
		this.rsaJsonWebKey = RsaJwkGenerator.generateJwk(2048);
		configureWithKey(consumer, claims, userCreator, this.rsaJsonWebKey.getKey());
		return this;
	}
	
	public JwtAuthenticator<U> configure(Consumer<JwtConsumerBuilder> consumer, BiConsumer<JwtClaims, U> claims, Function<JwtClaims, U> userCreator, Map<String, Object> keyParams) throws Exception{
		this.rsaJsonWebKey = new RsaJsonWebKey(keyParams);
		configureWithKey(consumer, claims, userCreator, this.rsaJsonWebKey.getKey());
		return this;
	}
	
	private JwtAuthenticator<U> configureWithKey(Consumer<JwtConsumerBuilder> consumer, BiConsumer<JwtClaims, U> claims, Function<JwtClaims, U> userCreator, Key key) throws Exception{
		JwtConsumerBuilder builder = new JwtConsumerBuilder()
			.setRequireExpirationTime()
			.setAllowedClockSkewInSeconds(30)
			.setVerificationKey(key)
			.setJwsAlgorithmConstraints(ConstraintType.PERMIT, AlgorithmIdentifiers.RSA_USING_SHA256);
				
			consumer.accept(builder);
				
			this.jwtConsumer = builder.build();
			this.claims = claims;
			this.userCreator = userCreator;
			
			return this;
	}
	
	public Map<String, Object> exportKey(){
		return this.rsaJsonWebKey.toParams(OutputControlLevel.INCLUDE_PRIVATE);
	}

	@Override
	public String createAuthUser(U user) throws Exception{
		JwtClaims claims = new JwtClaims();
		claims.setGeneratedJwtId(); // a unique identifier for the token
		claims.setIssuedAtToNow(); // when the token was issued/created (now)
		claims.setNotBeforeMinutesInThePast(2); // time before which the token is not yet valid (2 minutes ago)
		this.claims.accept(claims, user);

		JsonWebSignature jws = new JsonWebSignature();
		jws.setPayload(claims.toJson());
		jws.setKey(rsaJsonWebKey.getPrivateKey());
		jws.setKeyIdHeaderValue(rsaJsonWebKey.getKeyId());
		jws.setAlgorithmHeaderValue(AlgorithmIdentifiers.RSA_USING_SHA256);
		return jws.getCompactSerialization();
	}
	
	@Override
	public U getUser(RequestType type, HttpReader reader) throws Exception{
		String auth = reader.getHeader("Authorization");
		if(auth == null) return null;
		auth = auth.substring(7);
		return this.userCreator.apply(this.jwtConsumer.processToClaims(auth));
	}

	@Override
	public U getUser(String data) throws Exception {
		if(data == null) return null;
		return this.userCreator.apply(this.jwtConsumer.processToClaims(data));
	}

}
